from __future__ import absolute_import
from .asset import Asset


class FlightPhase(Asset):
	'''Retrieve the data of defined flight phases.'''

	def __init__(self, conn, ems_id=7):

		Asset.__init__(self, conn, "FlightPhase")
		self._ems_id = ems_id
		self.update_list()


	def update_list(self):

		Asset.update_list(self, uri_keys=('flt_phase','list'), uri_args=self._ems_id)
		self._rename_datacol('description', 'name')


	def get_id(self, name = None):

		a = self.search('name', name)['id'].tolist()
		return a if len(a) > 1 else a[0]


	def get_name(self, id_val = None):

		a = self.search('id', id_val, searchtype="match")['name'].tolist()
		return a if len(a) > 1 else a[0]


from builtins import object
from emspy.query import *


class Query(object):

	def __init__(self, conn, ems_name):

		self._conn 		= conn
		self._ems_name  = ems_name
		self.__ems 		= None
		self.__fleet 	= None
		self.__aircraft = None
		self.__airport 	= None
		self.__flightphase = None
		self.__init_ems()
		self._ems_id 	= self.__ems.get_id(ems_name)
		# self._init_assets()
		# self.reset()


	def __init_ems(self):

		self.__ems = EMS(self._conn)

	def get_ems_id(self):
		return self._ems_id

	def _init_assets(self):

		# self.__fleet = Fleet(self._conn, self._ems_id)
		# self.__aircraft = Aircraft(self._conn, self._ems_id)
		# self.__airport = Airport(self._conn, self._ems_id)
		# self.__flightphase = FlightPhase(self._conn, self._ems_id)		
		# self.__flight = Flight(self._conn, self._ems_id)
		raise NotImplementedError()


from __future__ import print_function
from builtins import object
from emspy.query import LocalData

import networkx as nx
import pandas as pd
import os, sys, re


class Flight(object):
    temp_exclude = ['Download Information', 'Download Review', 'Processing',
                    'Profile 16 Extra Data', 'Flight Review', 'Data Information',
                    'Operational Information', 'Operational Information (ODW2)',
                    'Weather Information', 'Profiles', 'Profile']


    def __init__(self, conn, ems_id, data_file=LocalData.default_data_file):

        self._conn = conn
        self._ems_id = ems_id
        self._db_id  = None
        self._metadata = None
        self._trees  = {'fieldtree': None, 'dbtree': None, 'kvmaps': None}
        self._fields = []
        self.__cntr = 0

        # Retreive the field tree data from local storage. If it doesn't exist, generate a new
        # default one.
        self.load_tree(data_file)
        # Set default database
        # self.set_database('fdw flight')

    def load_tree(self, file_name=LocalData.default_data_file):

        if self._metadata is None:
            self._metadata = LocalData(file_name)
        elif not self._metadata.is_db_path_correct(file_name):
            self._metadata.close()
            self._metadata = LocalData(file_name)

        self._trees = {'fieldtree': self.__get_fieldtree(),
                       'dbtree'   : self.__get_dbtree(),
                       'kvmaps'   : self.__get_kvmaps()}


    def save_tree(self, file_name=LocalData.default_data_file):

        from shutil import copyfile
        ld = self._metadata
        if not ld.is_db_path_correct(file_name):
            # A different location is given, if a previous file exists copy it to the new location.
            if ld.file_loc() != None and file_name != None:
                copyfile(ld.file_loc(), os.path.abspath(file_name))

            self._metadata.close()
            self._metadata = LocalData(file_name)

        self.__save_fieldtree()
        self.__save_dbtree()
        self.__save_kvmaps()



    def __get_fieldtree(self):
        if self._db_id is None:
            return pd.DataFrame(columns=self._metadata.table_info['fieldtree'])
        else:
            return self._metadata.get_data("fieldtree","ems_id = %d and db_id = '%s'" % (self._ems_id, self._db_id))


    def __save_fieldtree(self):
        if len(self._trees['fieldtree']) > 0:
            self._metadata.delete_data("fieldtree", "ems_id = %d and db_id = '%s'" % (self._ems_id, self._db_id))
            self._metadata.append_data("fieldtree", self._trees['fieldtree'])


    def __get_dbtree(self):
        T = self._metadata.get_data("dbtree", "ems_id = %d" % self._ems_id)
        if len(T) < 1:
            dbroot = {'ems_id': self._ems_id,
                      'id': "[-hub-][entity-type-group][[--][internal-type-group][root]]",
                      'name': "<root>",
                      'nodetype': "root",
                      'parent_id': None}
            self._trees['dbtree'] = pd.DataFrame([dbroot])
            self.__update_children(dbroot, treetype = "dbtree")
            self.update_tree("fdw", treetype = "dbtree", exclude_tree = ["APM Events"])
            self.__save_dbtree()
            T = self._trees['dbtree']
        return T


    def __save_dbtree(self):
        if len(self._trees['dbtree']) > 0:
            self._metadata.delete_data("dbtree", "ems_id = %d" % self._ems_id)
            self._metadata.append_data("dbtree", self._trees['dbtree'])


    def __get_kvmaps(self):
        T = self._metadata.get_data("kvmaps", "ems_id = %d" % self._ems_id)
        return T


    def __save_kvmaps(self):
        if len(self._trees['kvmaps']) > 0:
            self._metadata.delete_data("kvmaps", "ems_id = %d" % self._ems_id)
            self._metadata.append_data("kvmaps", self._trees['kvmaps'])


    def set_database(self, name):

        tr = self._trees['dbtree']
        self._db_id = tr[(tr.nodetype == 'database') & tr.name.str.contains(treat_spchar(name), case=False)]['id'].values[0]
        self._trees['fieldtree'] = self.__get_fieldtree()

        if self._trees['fieldtree'].empty:
            self.__update_children(self.get_database(), treetype = "fieldtree")

        print("Using database '%s'." % self.get_database()['name'])


    def get_database(self):

        tr = self._trees['dbtree']
        return tr[(tr.nodetype=="database") & (tr.id == self._db_id)].iloc[0].to_dict()



    def __db_request(self, parent):
        body = None
        if parent['nodetype'] == "database_group":
            body = {'groupId': parent['id']}
        resp_h, d = self._conn.request(uri_keys=('database', 'group'),
                                       uri_args=self._ems_id,
                                       body=body)
        d1 = []
        if len(d['databases']) > 0:
            d1 = [{'ems_id': parent['ems_id'],
                                'id': x['id'],
                                'nodetype': 'database',
                                'name': x['pluralName'],
                                'parent_id': parent['id']} for x in d['databases']]
        d2 = []
        if len(d['groups']) > 0:
            d2 = [{'ems_id': parent['ems_id'],
                                'id': x['id'],
                                'nodetype': 'database_group',
                                'name': x['name'],
                                'parent_id': parent['id']} for x in d['groups']]
        return d1, d2


    def __fl_request(self, parent):
        body = None
        if parent['nodetype'] == "field_group":
            body = {'groupId': parent['id']}
        resp_h, d = self._conn.request(uri_keys=('database', 'field_group'),
                                       uri_args=(self._ems_id, self._db_id),
                                       body=body)
        d1 = []
        if len(d['fields']) > 0:
            d1 = [{'ems_id': parent['ems_id'],
                                'db_id' : self._db_id,
                                'id': x['id'],
                                'nodetype': 'field',
                                'type': x['type'],
                                'name': x['name'],
                                'parent_id': parent['id']} for x in d['fields']]
        d2 = []
        if len(d['groups']) > 0:
            d2 = [{'ems_id': parent['ems_id'],
                                'db_id': self._db_id,
                                'id': x['id'],
                                'nodetype': 'field_group',
                                'type': None,
                                'name': x['name'],
                                'parent_id': parent['id']} for x in d['groups']]
        return d1, d2


    def __add_subtree(self, parent, exclude_tree=[], treetype = 'fieldtree'):

        print("On " + parent['name'] + "(" + parent['nodetype'] + ")" + "...")

        if treetype=="dbtree":
            searchtype = 'database'
            d1, d2 = self.__db_request(parent)

        else:
            searchtype = "field"
            d1, d2 = self.__fl_request(parent)

        if len(d1) > 0:
            self._trees[treetype] = self._trees[treetype].append(d1, ignore_index=True)
            plural = "s" if len(d1) > 1 else ""
            print("-- Added %d %s%s" % (len(d1), searchtype, plural))

        for x in d2:
            self._trees[treetype] = self._trees[treetype].append(x, ignore_index=True)
            if len(exclude_tree) > 0:
                if all([y not in x['name'] for y in exclude_tree]):
                    self.__add_subtree(x, exclude_tree, treetype)
            else:
                self.__add_subtree(x, exclude_tree, treetype)


    def __get_children(self, parent_id, treetype='fieldtree'):
        tr = self._trees[treetype]
        # tr = tr[tr.nodetype != ('field' if treetype=='fieldtree' else 'dbtree')]

        if isinstance(parent_id, (list, tuple, pd.Series)):
            return tr[tr.parent_id.isin(parent_id)]
        return tr[tr.parent_id == parent_id]


    def __remove_subtree(self, parent, treetype = 'fieldtree'):
        tr = self._trees[treetype]
        chld = tr[tr.parent_id == parent['id']]

        # Update the instance tree by deleting children
        self._trees[treetype] = tr[tr.parent_id != parent['id']]

        # Iterate and do recursive removal of children of children
        leaftype = 'field' if treetype=='fieldtree' else 'database'
        for i, x in chld[chld.nodetype!=leaftype].iterrows():
            self.__remove_subtree(x, treetype = treetype)


    # def __remove_subtree(self, parent, rm_parent=True, treetype='fieldtree'):

    #     rm_list = list()
    #     if rm_parent:
    #         rm_list.append(parent['id'])
    #     parent_id = [parent['id']]

    #     cntr = 0
    #     while len(parent_id) > 0:
    #         child_id = self.__get_children(parent_id, treetype=treetype)['id'].tolist()
    #         rm_list += child_id
    #         parent_id = child_id
    #         cntr += 1
    #         if cntr > 1e4:
    #             sys.exit("Something's wrong. Subtree removal went over 10,000 iterations.")
    #     if len(rm_list) > 0:
    #         tr = self._trees[treetype]
    #         self._trees[treetype] = tr[~tr.id.isin(rm_list)]
    #         print("Removed the subtree of %s (%s) with total of %d nodes (fields/databases/groups)." % (
    #             parent['name'], parent['nodetype'], len(rm_list)))



    def __update_children(self, parent, treetype='fieldtree'):
        '''
        This function updates the direct children of a parent node.
        '''
        print("On " + parent['name'] + "(" + parent['nodetype'] + ")" + "...")

        if treetype=="dbtree":
            searchtype = 'database'
            d1, d2 = self.__db_request(parent)

        else:
            searchtype = "field"
            d1, d2 = self.__fl_request(parent)

        T = self._trees[treetype]
        self._trees[treetype] = T[~((T.nodetype == searchtype) & (T.parent_id == parent['id']))]

        if len(d1) > 0:
            self._trees[treetype] = self._trees[treetype].append(d1, ignore_index=True)
            plural = "s" if len(d1) > 1 else ""
            print("-- Added %d %s%s" % (len(d1), searchtype, plural))

        # If there is an array of groups as children add any that appeared new and remove who does not.
        old_groups = T[(T.nodetype == '%s_group' % searchtype) & (T.parent_id == parent['id'])]
        old_ones = old_groups["id"].tolist()
        new_ones = [x['id'] for x in d2]

        rm_id = listdiff(old_ones, new_ones)
        if len(rm_id) > 0:
            [self.__remove_subtree(x.to_dict(), treetype=treetype) for i, x in old_groups.iterrows() if x['id'] in rm_id]

        add_id = listdiff(new_ones, old_ones)
        if len(add_id) > 0:
            self._trees[treetype] = self._trees[treetype].append([x for x in d2 if x['id'] in add_id])



    def update_tree(self, *args, **kwargs):
        '''
        Optional arguments
        ------------------

        treetype :
            "fieldtree" or "dbtree"

        exclude_tree:
            Exact name strings (case sensitive) of the field groups you don't want to search through
            Ex. ['Profiles', 'Weather Information']
        '''
        treetype     = kwargs.get("treetype", "fieldtree")
        exclude_tree = kwargs.get("exclude_tree", [])
        searchtype   = "field" if treetype == "fieldtree" else "database"

        if treetype not in ("fieldtree", "dbtree"):
            raise ValueError("treetype = '%s': there is no such data table." % treetype)

        fld_path = [s.lower() for s in args]

        for i, p in enumerate(fld_path):
            p = treat_spchar(p)
            if i == 0:
                T = self._trees[treetype]
                parent = T[T.name.str.contains(p, case=False)]
            else:
                self.__update_children(parent, treetype=treetype)
                chld_df = self.__get_children(parent['id'], treetype= treetype)
                child = chld_df[chld_df.name.str.contains(p, case=False)]
                parent = child
            if len(parent) == 0:
                raise ValueError("Search keyword '%s' did not return any %s group." % (p, searchtype))
            ptype  = "%s_group" % searchtype
            parent = parent[parent.nodetype == ptype]
            parent = get_shortest(parent)

        print("=== Starting to update subtree from '%s (%s)' ===" % (parent['name'], parent['nodetype']))
        self.__remove_subtree(parent, treetype=treetype)

        self.__add_subtree(parent, exclude_tree, treetype=treetype)



    def make_default_tree(self):
        dbnode = self.get_database()
        self.__remove_subtree(dbnode, treetype="fieldtree")
        self.__add_subtree(self.get_database(), exclude_tree=Flight.temp_exclude, treetype='fieldtree')
        # self.update_tree(self.get_database()['name'], exclude_tree=Flight.temp_exclude, treetype='fieldtree')


    def search_fields(self, *args, **kwargs):
        '''
        This function search through the field names and returns a list of tuples with
        (field_name, field_info). The input arg is a "keyword" for the field name. You can
        also add parent field group names as n-tuple.

        ex)
        ---
        search_fields("Takeoff Airport Code", "Landing Airport Code")
        search_fields(("Takeoff","Airport","Code"), ("Landing","Airport","Airport Code"))       '''

        unique = kwargs.get("unique", True)

        res = pd.DataFrame()

        for f in args:
            if type(f) is tuple and len(f) > 1:
                # If the given keyword is a tuple, search through the tree
                chld = self._trees['fieldtree']
                for i, ff in enumerate(f):
                    ff = treat_spchar(ff)
                    if i < (len(f)-1):
                        chld      = chld[chld.nodetype == "field_group"]
                        parent_id = chld[chld.name.str.contains(ff, case=False)]['id'].tolist()
                        tr        = self._trees['fieldtree']
                        chld      = tr[tr.parent_id.isin(parent_id)]
                    else:
                        chld = chld[(chld.nodetype == "field") & chld.name.str.contains(ff, case=False)]
                fres = chld
            else:
                # Simple keyword search
                T = self._trees['fieldtree']
                fres = T[(T.nodetype == "field") & T.name.str.contains(treat_spchar(f), case=False)]

            if fres.shape[0] == 0:
                # No returned value. Raise error.
                raise ValueError("No field found with field keyword %s." % f)
            elif fres.shape[0] > 1:
                if unique:
                    # If more than one value returned, choose one with the shortest name.
                    fres = get_shortest(fres)
            res = res.append(fres, ignore_index=True)

        # Convert the search result to a list of dicts
        res = [x[1].to_dict() for x in res.iterrows()]
        return res

    def list_allvalues(self, field=None, field_id=None, in_dict=False, in_df=False):
        '''
        List all available values for a discrete field. Will raise error if the type of
        a given field is not discrete.
        '''
        if field_id is None:
            fld = self.search_fields(field)[0]
            fld_type = fld['type']
            fld_id = fld['id']
            fld_name = fld['name']
            if fld_type != 'discrete':
                sys.exit("Queried field should be discrete to get the list of possible values.")
        else:
            fld_id = field_id
            tr     = self._trees['fieldtree']
            fld_name = tr.loc[tr.id == fld_id, 'name'].values[0]

        T = self._trees['kvmaps']
        kmap = T[(T.ems_id == self._ems_id) & (T.id == fld_id)]

        if len(kmap) == 0:
            print("%s: Getting key-value mappings from API. (Caution: Some fields take a very long time)" % fld_name)

            resp_h, content = self._conn.request(uri_keys=('database', 'field'),
                                                 uri_args=(self._ems_id, self._db_id, fld_id))
            km = content['discreteValues']
            kmap = pd.DataFrame({'ems_id': self._ems_id,
                                 'id': fld_id,
                                 'key': list(km.keys()),
                                 'value': list(km.values())})
            kmap['key'] = pd.to_numeric(kmap['key'])

            self._trees['kvmaps'] = self._trees['kvmaps'].append(kmap, ignore_index=True)
            self.__save_kvmaps()

        if in_dict:
            res = dict()
            for i, r in kmap.iterrows():
                res[r['key']] = r['value']
            return res

        if in_df:
            return kmap[['key', 'value']]
        return kmap['value'].tolist()

    def get_value_id(self, value, field=None, field_id=None):
        '''
        Return the key (Id) of the values of a discrete field.
        '''
        kvmap = self.list_allvalues(field=field, field_id=field_id, in_df=True)
        key = kvmap[kvmap.value == value]['key']

        if len(key) == 0:
            raise ValueError("%s could not be found from the list of the field values." % value)
        return int(key.values[0])


def get_shortest(fields):
    if isinstance(fields, pd.DataFrame) is False:
        sys.exit("Input should be a Pandas dataframe.")
    return fields.loc[fields.name.str.len().idxmin()].to_dict()


def listdiff(a, b):
    return [x for x in a if x not in b]


def treat_spchar(s):
    sp_chr = ("." ,"^" ,"(" ,")" ,"[" ,"]" ,"{", "}","<", ">","-", "+", "?", "!", "*", "$", "|", "&","%")
    for x in sp_chr:
        s = s.replace(x, "\\"+x)
    return s



from __future__ import print_function
from __future__ import absolute_import
from builtins import zip
from builtins import str
from emspy.query import *
from .query import Query

import pandas as pd
import sys, json


class FltQuery(Query):


	def __init__(self, conn, ems_name, data_file = LocalData.default_data_file):

		Query.__init__(self, conn, ems_name)
		self._init_assets(data_file)
		self.reset()


	def _init_assets(self, data_file):
		# Query._init_assets(self)
		self.__flight = Flight(self._conn, self._ems_id, data_file)


	def set_database(self, name):

		self.__flight.set_database(name)


	def get_database(self):

		return self.__flight.get_database()


	def reset(self):
		'''
		Resets the API query to send. This should be called before creating a new query.
		'''

		self.__columns = []

		self.__queryset = {
			"select": [],
			"groupBy": [],
			"orderBy": [],
			"distinct": True,
			"format": "none"
		}



	def select(self, *args, **kwargs):
		'''
		Functionally equivalent to SQL's select statement

		Example
		-------
		Following is the example of select method to query three fields and one more with
		aggregation function applied. The values are appended until the whole query is
		reset.

		>> query.select("customer id", "takeoff valid", "takeoff airport iata code")
		>> query.select("p301: fuel burned by all engines during cruise", aggregate="avg")
		'''
		aggs = ['none', 'avg', 'count', 'max', 'min', 'stdev', 'sum', 'var']
		aggregate = kwargs.get('aggregate', 'none')
		if aggregate not in aggs:
			sys.exit("Wrong aggregation selected. Use one of %s." % aggs)
		fields = self.__flight.search_fields(*args, **kwargs)

		if type(fields)!=list: fields = [fields]

		for field in fields:
			d = {}
			d['fieldId'] = field['id']
			d['aggregate'] = aggregate
			self.__queryset['select'].append(d)
			self.__columns.append(field)


	def group_by(self, *args):
		'''Functionally equivalent to SQL's groupby'''
		for field in self.__flight.search_fields(*args):
			self.__queryset['groupBy'].append({'fieldId': field['id']})


	def order_by(self, field, order='asc'):
		'''Functional equivalent of SQL's order by'''
		if order not in ['asc', 'desc']:
			sys.exit("Ordering option must be one of %s." % ['asc', 'desc'])
		self.__queryset['orderBy'].append({
			'fieldId': self.__flight.search_fields(field)[0]['id'],
			'order': order,
			'aggregate': 'none'})


	def filter(self, expr):
		'''
		Translate the give filtering conditions to json queries.
		'''
		if 'filter' not in self.__queryset:
			self.__queryset['filter'] = {
				'operator': 'and',
				'args': []
			}
		expr_vec = self.__split_expr(expr)
		jsonobj  = self.__translate_expr(expr_vec)
		self.__queryset['filter']['args'].append(jsonobj)



	def __split_expr(self, expr):
		'''
		This function might need to be updated to be more robust. The split of a
		given expression will not be correctly done if either field or field
		value contains symbols identical to filtering operators.

		In fact the correction was already applied in Rems so refer to Rems'
		split_expr function to check what was done there.
		'''
		import re

		for pattern in ['[=!<>]=?'] + list(sp_ops.keys()):
			a = re.search(pattern, expr)
			if a is not None:
				break

		if a is None:
			sys.exit("Cannot find any valid conditional operator from the given string expression.")
		splitted = expr.partition(a.group())
		return splitted


	def __translate_expr(self, expr_vec):

		fld_loc = [False, False]
		fld_info = None
		fld_type = None
		val_info = None
		op = expr_vec[1]

		for i, s in enumerate([expr_vec[0], expr_vec[2]]):
			x = eval(s)

			if i == 0:
				fld = self.__flight.search_fields(x)[0]
				if fld is not None:
					fld_info = [{'type':'field', 'value': fld['id']}]
					fld_type = fld['type']
					fld_loc[i] = True
				else:
					raise ValueError("No field was found with the keyword %s. Please double-check if it is a right keyword." % x)
			else:
				if type(x) != list:
					x = [x]
				val_info = [{'type':'constant','value':v} for v in x]

		if fld_loc[1]:
			if '<' in expr_vec[1]:
				op = expr_vec[1].replace('<','>')
			else:
				op = expr_vec[1].replace('>','<')

		arg_list = fld_info + val_info

		if fld_type=="boolean":
			fltr = _boolean_filter(op, arg_list)
		elif fld_type=="discrete":
			fltr = _discrete_filter(op, arg_list, self.__flight)
		elif fld_type=="number":
			fltr = _number_filter(op, arg_list)
		elif fld_type=="string":
			fltr = _string_filter(op, arg_list)
		elif fld_type=="dateTime":
			fltr = _datetime_filter(op, arg_list)
		else:
			raise ValueError("%s has an unknown field data type %s." % (fld[0], fld_type))
		return fltr


	def distinct(self, x=True):

		self.__queryset['distinct'] = x


	def get_top(self, n):

		self.__queryset['top'] = n


	# def readable_output(self, x=False):

	# 	if x:
	# 		y = "display"
	# 	else:
	# 		y = "none"
	# 	self.__queryset['format'] = y


	def in_json(self):

		return json.dumps(self.__queryset)


	def in_dict(self):

		return self.__queryset


	def simple_run(self, output = "dataframe"):
		'''
		Sends query to EMS API via the regular query call. The regular query call has a size limit
		in the returned data, which is 25000 rows max. Any output that has greater than 25000 rows
		will be truncated. For the query that is expected to return with large data. Please use the
		async_run method.

		Input
		-----
		output: desired output data format. Either "raw" or "dataframe".

		Output
		------
		Returned data for query in Pandas' DataFrame format
		'''
		print('Sending a simple query to EMS ...')
		resp_h, content = self._conn.request(
			rtype="POST",
			uri_keys=('database','query'),
			uri_args=(self._ems_id, self.__flight.get_database()['id']),
			jsondata= self.__queryset
			)
		print('Done.')

		if output == "raw":
			return content
		elif output == "dataframe":
			return self.__to_dataframe(content)
		else:
			raise ValueError("Requested an unknown output type.")



	def async_run(self, n_row = 25000):
		'''
		Sends query to EMS API via async-query call. The async-query does not process
		the query as a single batch for a query expecting a large data. You will have
		to call it multiple times. This function do this multiple calls for you.

		Input
		-----
		n_row: batch size of a single async call. Default is 25000.

		Output
		------
		Returned data for query in Pandas' DataFrame format
		'''

		print('Sending and opening an async-query to EMS ...', end=' ')
		resp_h, content = self._conn.request(
			rtype = "POST",
			uri_keys = ('database', 'open_asyncq'),
			uri_args = (self._ems_id, self.__flight.get_database()['id']),
			jsondata = self.__queryset
			)
		if 'id' not in content:
			sys.exit("Opening Async query did not return the query Id.")
		query_id = content['id']
		query_header = content['header']
		print('Done.')

		ctr = 0
		df = None
		while True:
			print(" === Async call: %d ===" % (ctr+1))
			try:
				resp_h, content = self._conn.request(
					rtype= "GET",
					uri_keys = ('database', 'get_asyncq'),
					uri_args = (self._ems_id,
								self.__flight.get_database()['id'],
								query_id,
								n_row*ctr,
								n_row*(ctr+1)-1)
					)
				content['header'] = query_header
				dff = self.__to_dataframe(content)
			except:
				print("Something's wrong. Returning what has been sent so far.")
				# from pprint import pprint
				# pprint(resp_h)
				# pprint(content)
				return df

			if ctr == 0:
				df = dff
			else:
				df = df.append(dff, ignore_index = True)

			print("Received up to %d rows." % df.shape[0])
			if dff.shape[0] < n_row:
				break
			ctr += 1

		print("Done.")
		return df


	def run(self, n_row = 25000):
		'''
		Sends query to EMS API. It uses either regular or async query call depending on
		the expected size of output data. It supports only Pandas DataFrame as the output
		format.

		Input
		-----
		n_row: batch size of a single async call. Default is 25000.

		Output
		------
		Returned data for query in Pandas' DataFrame format
		'''
		Nout = None
		if 'top' in self.__queryset:
			Nout = self.__queryset['top']

		if (Nout is not None) and (Nout <= 25000):
			return self.simple_run(output= "dataframe")

		return self.async_run(n_row = n_row)


	def __to_dataframe(self, json_output):
		'''
		Changes Dict (JSON) formatted raw output from the EMS API to Pandas'
		DataFrame.
		'''

		print("Raw JSON output to Pandas dataframe...")
		col      = [h['name'] for h in json_output['header']]
		coltypes = [c['type'] for c in self.__columns]
		col_id   = [c['id'] for c in self.__columns]
		val      = json_output['rows']

		df = pd.DataFrame(data = val, columns = col)

		if df.empty: return df

		if self.__queryset['format'] == "display":
			print("Done.")
			return df

		# Do the dirty work of casting a right type for each column of the data
		# Note
		# ====
		# Runway IDs are discrete data but their key-value mapping is not provided
		# because the mapping itself is quite big in size (45K entries). That means
		# the regular routines to handle the discrete data won't work. As a result
		# the discrete data routine has a dirty, custom routine particularly for
		# the runway IDs. What it basically does is to send a separate but redundant
		# query for runway IDs with "queryset$format = display", and then push the
		# this query result at the runway ID column of the original query result.
		# I know this is crappy but it seems the best way I could find.
		for i, cid, cname, ctype in zip(range(len(col)), col_id, col, coltypes):
			try:
				if ctype=='number':
					df.iloc[:, i] = pd.to_numeric(df.iloc[:, i])
				elif ctype=='discrete':
					df.iloc[:, i] = self.__key_to_val(df.iloc[:, i], cid)
					# k_map = self.__flight.list_allvalues(field_id = cid, in_dict = True)
					# if len(k_map) == 0:
					# 	df[cname] = self.__get_rwy_id(cname)
					# else:
					# 	df = df.replace({cname: k_map})
				elif ctype=='boolean':
					df.iloc[:, i] = df.iloc[:, i].astype(bool)
				elif ctype=='dateTime':
					df.iloc[:, i] = pd.to_datetime(df.iloc[:, i]).dt.tz_localize('UTC')
			except ValueError:
				print("Somethings wrong when converting to Pandas DataFrame for column '%s' (type: %s)." % (cname, ctype))
		print("Done.")
		return df


	def __key_to_val(self, ds, field_id):

		k_map = self.__flight.list_allvalues(field_id = field_id, in_df = True)

		# Sometimes k_map is a very large table making "replace" operation
		# very slow. Just grap kv-maps subset that are present in the target
		# dataframe
		k_map = k_map[k_map.key.isin(ds.unique())]
		# Change k_map in dict
		km_dict = dict()
		for i, r in k_map.iterrows():
			km_dict[r['key']] = r['value']
		ds = ds.replace(km_dict)

		return ds


	def __get_rwy_id(self, cname):
		'''
		Deprecated
		'''

		print("\n --Running a special routine for querying runway IDs. This will make the querying twice longer.")
		qs = self.__queryset
		qs['format'] = "display"

		resp_h, content = self._conn.request(
			rtype="POST",
			uri_keys=('database','query'),
			uri_args=(self._ems_id, self.__flight.get_database()['id']),
			jsondata= qs
			)
		return self.__to_dataframe(content)[cname]


	def update_dbtree(self, *args, **kwargs):
		exclude_tree = kwargs.get("exclude_tree",[])
		self.__flight.update_tree(*args, **{"treetype":"dbtree", "exclude_tree":exclude_tree})


	def update_fieldtree(self, *args, **kwargs):
		exclude_tree = kwargs.get("exclude_tree", [])
		self.__flight.update_tree(*args, **{"treetype":"fieldtree", "exclude_tree":exclude_tree})

	def generate_preset_fieldtree(self):
		self.__flight.make_default_tree()


	def save_metadata(self, file_name = None):
		self.__flight.save_tree(file_name)


	def load_metadata(self, file_name = None):
		self.__flight.load_tree(file_name)









## Experimental...

basic_ops = {
	'==': 'equal', '!=': 'notEqual', '<': 'lessThan',
	'<=': 'lessThanOrEqual', '>': 'greaterThan',
	'>=': 'greaterThanOrEqual'
}
sp_ops = {
	' in ': 'in',
	' not in ': 'notIn'
}
# '=Null': 'isNull', '!=Null': 'isNotNull', 'and': 'And', 'or': 'Or', 'in': 'in', 'not in': 'notIn'


def _filter_fmt1(op, *args):
	fltr = {
		"type": "filter",
		"value": {
			"operator": op,
			"args": []
		}
	}
	for x in args:
		fltr['value']['args'].append({'type': x['type'], 'value': x['value']})

	return fltr


def _boolean_filter(op, d):

	fld_info = d[0]
	val_info = d[1]
	if type(val_info['value'])!=bool: raise ValueError("%s: use a boolean value." % val_info['value'])
	if op == "==":
		t_op = 'is'+str(val_info['value'])
	elif op == "!=":
		t_op = 'is'+str(not val_info['value'])
	else:
		raise ValueError("Conditional operator %s is given. Booleans shoule be only with boolean operators." % op)

	fltr = _filter_fmt1(t_op, fld_info)
	return fltr


def _discrete_filter(op, d, flt):
	fld_info = d[0]

	if op in list(basic_ops.keys()):
		# Single input basic coniditonal operation
		t_op = basic_ops[op]
		val_info = d[1]
		vid  = flt.get_value_id(val_info['value'], field_id = fld_info['value'])
		val_info['value'] = vid
		fltr = _filter_fmt1(t_op, fld_info, val_info)

	elif op in [" in ", " not in "]:
		t_op = sp_ops[op]
		val_list = [{'type':x['type'], 'value':flt.get_value_id(x['value'], field_id=fld_info['value'])}\
					 for x in d[1:]]
		inp = [fld_info] + val_list
		fltr = _filter_fmt1(t_op, *inp)
	else:
		raise ValueError("%s: Unsupported conditional operator for discrete field type." % op)
	return fltr


def _number_filter(op, d):

	if op in basic_ops:
		t_op = basic_ops[op]
		fltr = _filter_fmt1(t_op, d[0], d[1])
	else:
		raise ValueError("%s: Unsupported conditional operator for number field type." % op)
	return fltr


def _string_filter(op, d):

	if op in ["==", "!="]:
		t_op = basic_ops[op]
		fltr = _filter_fmt1(t_op, d[0], d[1])

	elif op in [" in ", " not in "]:
		t_op = sp_ops[op]
		fltr = _filter_fmt1(t_op, *d)

	else:
		raise ValueError("%s: Unsupported conditional operator for string field type." % op)
	return fltr


def _datetime_filter(op, d):

	from datetime import datetime

	date_ops = {
		"<": "dateTimeBefore",
		">=": "dateTimeOnAfter"
	}

	if op in list(date_ops.keys()):
		t_op = date_ops[op]
		fltr = _filter_fmt1(t_op, d[0], d[1])
		# Additional json attribute to specify this is UTC time
		fltr['value']['args'].append({'type':'constant', 'value': 'Utc'})
	else:
		raise ValueError("%s: Unsupported conditional operator for datetime field type." % op)
	return fltr









from __future__ import print_function
from builtins import range
from builtins import object
import pandas as pd
import numpy as np
import os, sys, re
import emspy
from emspy.query import LocalData

class Analytic(object):

	def __init__(self, conn, ems_id, data_file = LocalData.default_data_file):
		self._conn        = conn
		self._ems_id      = ems_id
		self._metadata    = None
		self._load_paramtable(data_file)


	def _load_paramtable(self, file_name = LocalData.default_data_file):
		if self._metadata is None:
			self._metadata = LocalData(file_name)
		elif not self._metadata.is_db_path_correct(file_name):
			self._metadata.close()
			self._metadata = LocalData(file_name)

		self._param_table = self._metadata.get_data("params", "ems_id = %d" % self._ems_id)

	
	def _save_paramtable(self):
		if len(self._param_table) > 0:
			self._metadata.delete_data("params", "ems_id = %d" % self._ems_id)
			self._metadata.append_data("params", self._param_table)


	def get_param_details(self, analytic_id):
		resp_h, content = self._conn.request(rtype="POST", uri_keys=('analytic', 'search'), uri_args=self._ems_id,
											 jsondata={'id': analytic_id})

		return content

	def search_param(self, keyword, in_df = False):
		print('Searching for params with keyword "%s" from EMS ...' % keyword, end=' ')
		# EMS API Call
		resp_h, content = self._conn.request( uri_keys=('analytic', 'search'),
											  uri_args=self._ems_id,
											  body={'text': keyword})
		if len(content) == 0:
			sys.exit("No parameter found with search keyword %s." % keyword)
		elif len(content) == 1:
			res = content
		else:
			word_len     = [len(x['name']) for x in content]
			idx          = np.argsort(word_len).tolist()
			res          = [content[i] for i in idx]
		print("done.")

		for i in range(len(res)):
			res[i]['ems_id'] = self._ems_id

		if in_df:
			return pd.DataFrame(res)
		
		return res
		

	def get_param(self, keyword, unique = True):
		# if the param table is empty, just return an empty param dict.
		if self._param_table.empty:
			return dict(ems_id="", id="", name="", description="", units="")
		# If the param table is not empty, do search by keyword
		bool_idx = self._param_table['name'].str.contains(keyword, case = False, regex=False)
		df = self._param_table[bool_idx]
		# If the search result is empty, return empty param dict
		if df.empty:
			return dict(ems_id="", id="", name="", description="", units="")
		# If not empty, return the one with shortest name
		if df.shape[0] > 1:
			idx = df['name'].map(lambda x: len(x)).sort_values().index
			df  = df.loc[idx, :]
		# When unique = True
		if unique:
			return df.iloc[0,:].to_dict()
		# When unique = False
		return df.to_dict('records')







from .query import Query
from .ems import EMS
import pandas as pd
import json


class InsertQuery(Query):
    """
    A class for performing database inserts using the EMS API.

    Args:
        conn (ems.connection): A valid instantiated EMS connection object.
        ems_name (str): A valid EMS system name.
        database_id (str): A database id to insert into.
    """

    def __init__(self, conn, ems_name, database_id):
        Query.__init__(self, conn, ems_name)
        self._database_id = database_id
        self.__create_columns = {'createColumns': []}

    def in_json(self):
        return json.dumps(self.__create_columns)

    def reset(self):
        '''
        Resets the API query to send. This should be called before creating a new query.
        '''

        self.__create_columns = {'createColumns': []}

    def get_create_columns(self):
        """
        A method which returns the values of createColumns attribute.

        Returns:
            dictionary: a dictionary corresponding to the current value of self.__create_columns
        """

        return self.__create_columns

    def insert_df(self, df, schema_map=None):
        """
        A method for insert values from a pandas DataFrame.

        Args:
            df (:obj:`pd.DataFrame`): A pandas DataFrame of values to input, where the columns are the fieldIds and the
                entries are values to input (unless schema_map is passed, in which case column names can be arbitrary
                as long as they exist in schema_map and map to ems monikers).
            schema_map (:obj:`dict`, optional): A mapping of named DataFrame columns to field ids, e.g.
                {'column1' = '[-hub][schema]'}.  If this is not passed, the columns of df should correspond to EMS
                schemas.
        Raises:
            TypeError: If `schema_map` is not a dictionary.
            TypeError: If `df` is not a pandas DataFrame.
            ValueError: If `schema_map` is passed, but not all `df` column names are present in `schema_map`.
        """

        if not isinstance(df, pd.DataFrame):
            raise TypeError("`df` must be a pandas DataFarme.  Found: {0}".format(type(df)))

        # If a schema_map was passed in (mapping column name to ems schema)
        if schema_map is not None:
            # schema_map should be a dictionary.
            if not isinstance(schema_map, dict):
                raise TypeError("`schema_map` must be a dictionary.  Found: {0}".format(type(df)))
            # Find columns that exist in `df` but not in `schema_map`.  Convert to list.
            missing_from_map = list(set(df.columns.values) - set(schema_map.keys()))
            # If any columns are in `df` but not in `schema_map`, error out.
            if len(missing_from_map) > 0:
                raise ValueError("Column(s): '{0}' found in `df`, but not in `schema_map`.".format(missing_from_map))
            df = df.rename(columns=schema_map)
        else:
            print('schema_map was not passed.  Assuming dataframe column names correspond to EMS schema.')

        for idx, row in df.iterrows():
            row_dict = row.to_dict()
            row_list = [{'fieldId': fid, 'value': val} for fid, val in zip(row_dict.keys(), row_dict.values())]
            self.__create_columns['createColumns'].append(row_list)

    def run(self):
        resp_h, content = self._conn.request(rtype='POST', uri_keys=('database', 'create'),
                                             uri_args=(self._ems_id, self._database_id), jsondata=self.__create_columns)
        # Unintended responses are handled in Connection.request, so we only do verification here.
        rows_added = content['rowsAdded']
        if (rows_added == len(self.__create_columns['createColumns'])):
            print('Added all rows.')



from __future__ import absolute_import
from .asset import Asset


class Airport(Asset):
	'''Manages airport info'''
	_col_tr = {
		'default': 'codePreferred',
		'faa': 'codeFaa',
		'icao': 'codeIcao',
		'city': 'city',
		'name': 'name'
	}

	def __init__(self, conn, ems_id):

		Asset.__init__(self, conn, "Airport")
		self._ems_id = ems_id
		self.update_list()


	def update_list(self):

		Asset.update_list(self, uri_keys=('airport','list'), uri_args=self._ems_id, colsort=False)


	def get_id(self, name = None, code='default'):

		a = self.search(Airport._col_tr[code], name)['id'].tolist()
		return a if len(a) > 1 else a[0]


	def get_name(self, id_val = None, code='default'):

		a = self.search('id', id_val, searchtype="match")[Airport._col_tr[code]].tolist()
		return a if len(a) > 1 else a[0]


from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import map
from builtins import object
import json, urllib.request, urllib.parse, urllib.error, urllib.request, urllib.error, urllib.parse, ssl, sys, io, gzip
from numbers import Number
import pprint as pp
from . import common



class Connection(object):
	'''
	Object for connection to EMS API
	'''
	def __init__(self, user=None, pwd=None, proxies=None, verbose=False, ignore_ssl_errors=False, server="prod", server_url=None):

		self.__user 		= user
		self.__pwd  		= pwd
		self.__proxies      = proxies
		self.__ntrials      = 0
		self.__uri_root     = None
		self.__ignore_ssl_errors = ignore_ssl_errors
		self.token 			= None
		self.token_type 	= None

		# We assign the uri root to a member variable up front, and use that everywhere to
		# simplify. In order to use an alternate uri root, it must be specified in the constructor.
		if server_url is not None:
			self.__uri_root = server_url
		else:
			self.__uri_root = common.uri_root[server]

		if (user is not None) and (pwd is not None):
			self.connect(user, pwd, proxies, verbose)
		else:
			print("An empty connection is instantiated because credentials are not provided.\n")



	def connect(self, user, pwd, proxies=None, verbose=False):
		'''
		Connect to EMS system using given credentials.
		'''
		if proxies is not None:
			proxy_handler = urllib.request.ProxyHandler(proxies)
			opener = urllib.request.build_opener(proxy_handler, urllib.request.HTTPHandler)
			urllib.request.install_opener(opener)	

		headers = {'Content-Type':'application/x-www-form-urlencoded', 'User-Agent':common.user_agent}
		data   = {'grant_type': 'password', 'username': user, 'password': pwd}

		resp_h, content = self.request(
			rtype="POST", uri_keys=('sys','auth'), data=data, 
			headers = headers, proxies=proxies, verbose=verbose
			)

		# Add error handling --

		# Get the token
		self.token      = content['access_token']
		self.token_type = content['token_type']
		return resp_h, content


	def reconnect(self, verbose = False):

		if self.__ntrials >= 3:
			sys.exit("Stop trying to reconnect EMS API after %d trials" % self.__ntrials)

		self.__ntrials +=1
		return self.connect(self.__user, self.__pwd, self.__proxies, verbose)


	def request(self,
			rtype="GET", uri=None, uri_keys=None, uri_args=None, 
			headers=None, body=None, data=None, jsondata=None, proxies=None, 
			verbose=False
		):

		# If no custom headers are given, use our own
		if headers is None: 
			headers = {'Authorization': ' '.join([self.token_type, self.token]), 'Accept-Encoding': 'gzip', 'User-Agent': common.user_agent }

		# If uri_keys are given, find the uri from the uris dictionary
		if uri_keys is not None:
			uri    = self.__uri_root + common.uris[uri_keys[0]][uri_keys[1]]

		# Provide the input to the request
		if uri_args is not None:
			# uri    = uri % uri_args
			# Unencoded url does not work all of sudden...
			
			# encode_args = lambda x: urllib.parse.quote(x) if type(x) in (str, unicode) else x
			# if type(uri_args) in (list, tuple):
			# 	uri = uri % tuple(map(encode_args, uri_args))
			# else:
			# 	uri = uri % encode_args(uri_args)
			
			if type(uri_args) not in (list, tuple):
				uri_args = [uri_args]
				
			uri_args = [x if isinstance(x, Number) else urllib.parse.quote(x) \
						for x in uri_args]
			uri = uri % tuple(uri_args)
			
		# Append query to the uri if body is given
		if body is not None:
			uri    = uri + "?" + urllib.parse.urlencode(body)

		# Encode the data
		if data is not None:
			data 	= urllib.parse.urlencode(data).encode('utf-8')

		if jsondata is not None:
			headers['Content-Type'] = 'application/json'
			data = json.dumps(jsondata).encode('utf-8')

		# uri = uri.encode('utf-8')
		req = urllib.request.Request(uri, data=data, headers=headers)
		try:
			resp = self.__send_request(req)
			statcode = resp.getcode()
			if statcode!=200:
				print("Http status code: %d" % statcode)
				verbose = True
		except:
			(eType, eValue, eTrace) = sys.exc_info()
			if eType is ssl.CertificateError:
				print("A certificate verification error occured for the request to '%s'. Certificate verification is required by default, but can be disabled by using the ignore_ssl_errors argument for the Connection constructor." % uri )
				raise
			if eType is urllib.error.HTTPError:
				message_bytes = eValue.read()
				message_str = message_bytes.decode('utf-8')
				eValue.msg = eValue.msg + '\n Details: ' + message_str
				raise

			print("Trying to reconnect the EMS API.")
			self.reconnect()
			print("Done.")
			resp = self.__send_request(req)
		resp_h   = resp.getheaders()

		# If the response is compressed, decompress it.
		if resp.info().get('Content-Encoding') == 'gzip':
			buffer = io.BytesIO(resp.read())
			file = gzip.GzipFile(fileobj=buffer)
			content = json.loads(file.read())
		else:
			content = json.loads(resp.read())
			
		if verbose:
			print("URL: %s" % resp.geturl())
			pp.pprint(resp_h)
			pp.pprint(content)

		return resp_h, content


	def __send_request(self, req):
		"""Sends the request and returns the response, optionally ignoring ssl errors."""

		# Normally you do NOT want to ignore SSL errors, but this is
		# sometimes necessary on beta API endpoints without a proper cert.
		# TODO: Find a real fix for GE Mac OS machines that are having trouble verifying.  
		request = None
		if self.__ignore_ssl_errors:
			try:
				request = urllib.request.urlopen(req, context=None)
			except urllib.error.URLError:
				request = urllib.request.urlopen(req, context=ssl._create_unverified_context())
		else:
			request = urllib.request.urlopen(req, context=None)
		return request


def print_resp(resp):
	
	for r in resp:
		pp.pprint(r)



import pandas as pd
from .query import Query


class Profile(Query):
    """
    A class for performing profile queries in the EMS API.

    Args:
        conn (ems.connection): A valid instantiated EMS connection object.
        ems_name (str): A valid EMS system name.
        profile_number (:obj: `int`, optional): A profile number to select.
        profile_name (:obj: `str`, optional): A profile name.

    Raises:
        ValueError: If neither `profile_number` and `profile_name` are specified.
        LookupError: If a profile matching either the `profile_number` or `profile_name` is not found.
    """

    def __init__(self, conn, ems_name, profile_number=None, profile_name=''):
        Query.__init__(self, conn, ems_name)

        # Set parameters from arguments
        self._conn = conn
        self._ems_name = ems_name
        self._input_profile_number = profile_number
        self._input_profile_name = profile_name

        # To be filled out later
        self._guid = None
        self._glossary = None
        self._events_glossary = None
        self._exact_search = False

        # Get EMS id.
        self._ems_id = self.get_ems_id()

        # Set up template strings.
        self._no_matches_name_template = "Did not find any profiles that could match profile_name: {0}."
        self._no_matches_num_template = "Did not find any profiles that could match profile_number: {0}."
        self._many_matches_name_template = "Found multiple profiles that could match {0}:\n\n {1}"
        self._many_matches_num_template = "Found multiple profiles that could match {0}:\n\n {1}"

        # Check that at least one of the profile number/name arguments was passed.
        if profile_number is None and profile_name == '':
            raise ValueError("Either profile_number, profile_name, or both must be specified.")

        # Decide how to search
        self._search_type = 'number' if profile_number is not None else 'name'
        # Perform a search using the specified options
        self._search(self._search_type)

    def get_glossary(self):
        """
        Returns:
            A Pandas DataFrame containing a profile glossary.
        """

        if self._guid is not None:
            resp_h, dict_data = self._conn.request(uri_keys=('profile', 'glossary'),
                                                   uri_args=(self._ems_id, self._guid))
            a = pd.DataFrame.from_dict(dict_data)
            # convert the dictionaries values within the glossaryItems column of a into a new DataFrame
            b = a['glossaryItems'].apply(pd.Series)
            # concatenate a and b, drop the old glossaryItems column name
            c = pd.concat([a, b], axis=1).drop('glossaryItems', axis=1)
            self._glossary = c
            return self._glossary
        else:
            print("The search results did not return a profile matching the input profile name and number on the given"
                  "system.  Please try to instantiate the profile object with different arguments.")
            return

    def get_events_glossary(self):
        if self._guid is not None:
            resp_h, dict_data = self._conn.request(uri_keys=('profile', 'events'),
                                                   uri_args=(self._ems_id, self._guid))
            a = pd.DataFrame(dict_data)
            a.set_index('id', inplace=True)
            self._events_glossary = a
            return self._events_glossary
        else:
            print("The search results did not return a profile matching the input profile name and number on the given"
                  "system.  Please try to instantiate the profile object with different arguments.")
            return

    def get_profile_results(self, flight_id):
        raise NotImplementedError
        if self._guid is not None:
            resp_h, dict_data = self._conn.request(uri_keys=('profile', 'profile_results'),
                                                   uri_args=(self._ems_id, flight_id, self._guid))
            '''
            TODO: Implement some sort of parsing of resulting data.  Honestly, it's not super useful unless paired
            with some supplementary information.  There are no names.  No descriptions.  Perhaps I could combine this
            information with the information obtained in the "glossary" and create a meaningful output.   
            
            Glossary, however, does not contain the names and ID's of events.  Maybe those will have to be grabbed using
            the other API endpoints.  
            '''
        else:
            print("The search results did not return a profile matching the input profile name and number on the given"
                  "system.  Please try to instantiate the profile object with different arguments.")
            return

    def __set_profile_attributes(self, profile_df):
        self._guid = profile_df['id'].values[0]
        self._current_version = profile_df['currentVersion'].values[0]
        self._library = profile_df['library'].values[0]
        self._profile_name = profile_df['name'].values[0]
        self._local_id = profile_df['localId'].values[0]

    def __request_results(self):
        # get values via API.
        resp_h, dict_data = self._conn.request(uri_keys=('profile', 'search'), uri_args=self._ems_id,
                                               body={'search': self._input_profile_name})
        a = pd.DataFrame.from_dict(dict_data)  # create a DataFrame from the response.
        self._search_results = a
        return a

    def __validate_search_results(self, res):
        if len(res) == 0:
            raise LookupError(self._no_matches_name_template.format(self._input_profile_name))

    def __filter_results(self, res):
        filtered = res  # default assignment
        # search can be of three different types: {name+exact, name+inexact, number}
        if self._search_type == 'name':
            if self._exact_search is True:  # name + exact search
                # locate profiles using exact string match.
                filtered = res.loc[res['name'] == self._input_profile_name, :]
            else:  # name + inexact search
                # locate profiles with string in name (case insensitive).
                res['name_lower'] = res['name'].str.lower()
                filtered = res.loc[res['name_lower'].str.contains(self._input_profile_name.lower()), :]
        elif self._search_type == 'number':  # number search
            # locate rows where the localId matches the input profile number.
            filtered = res.loc[res['localId'] == self._input_profile_number]
        return filtered

    def __validate_filtered(self, filtered):
        # Check the filtered values.
        # If only one row is found, set the class attributes using this row.
        if len(filtered) == 1:
            print('Found a profile with the supplied profile number and name.')
            print('Profile name: {0}, profile number: {1}.'.format(filtered['name'].iloc[0],
                                                                   filtered['localId'].iloc[0]))
            self.__set_profile_attributes(filtered)
        # if more than one rows are found (shouldn't be possible) either try again (if we are searching by name in an
        # inexact manner, we will now try an exact search) or return a LookupError.
        # Inexact search could lead to more than one match, leaving the potential for an exact search to narrow to one.
        # It should not be possible to have more than one match if you are searching with a profile number.
        elif len(filtered) > 1:
            match_names = filtered['name'].values
            match_profiles = filtered['localId'].values
            match_strings = ['P' + str(pnum) + ': ' + name for pnum, name in zip(match_profiles, match_names)]
            match_names_formatted = '\t' + '\n\t'.join(match_strings)  # format a string for printing
            print(self._many_matches_name_template.format(self._input_profile_name, match_names_formatted))
            if self._exact_search is False:
                print('Attempting exact string match.')
                self._exact_search = True
                filtered = self.__filter_results(filtered)  # try using exact string matching to pare down results.
                self.__validate_filtered(filtered)  # validate again, using newly obtained exact match results.
            else:
                raise LookupError(self._many_matches_name_template.format(self._input_profile_name,
                                                                          match_names_formatted))
        # if no rows are found, return an error.
        elif len(filtered) == 0:
            # choose the right template for the returned error.
            if self._search_type == 'number':
                raise LookupError(self._no_matches_num_template.format(self._input_profile_number))
            elif self._search_type == 'name':
                raise LookupError(self._no_matches_name_template.format(self._input_profile_name))

    def _search(self, search_type):
        res = self.__request_results()
        self.__validate_search_results(res)
        filtered = self.__filter_results(res)
        self.__validate_filtered(filtered)




from __future__ import absolute_import
from .asset import Asset


class EMS(Asset):
	'''Retrieves EMS system-specific info'''

	def __init__(self, conn):
		Asset.__init__(self, conn, "EMS")


	def update_list(self):
		Asset.update_list(self, uri_keys=('ems_sys', 'list'))


	def ensure_loaded(self):
		if not (Asset.list_all(self)):
			self.update_list()


	def get_id(self, name = None):
		"""
		A method to get an ems_id using a supplied ems_name.

		Args:
		    ems_name (str): An EMS system name to get the ID for.
		Raises:
		    ValueError: If `ems_name` is not specified, but the user has access to more than one system.
		"""

		# Support using integer IDs directly
		if isinstance(name, int):
			return name

		self.ensure_loaded()
		name = name.upper()
		a    = self.search('name', name, searchtype="match")['id'].tolist()
		if len(a) == 0:
			sys_names = self.list_all()['name'].to_list()
			raise ValueError('No matching systems found.  You have access to: {0}'.format(sys_names))

		return a if len(a) > 1 else a[0]


	def get_name(self, id_val = None):
		self.ensure_loaded()
		a = self.search('id', id_val, searchtype="match")['name'].tolist()
		return a if len(a) > 1 else a[0]





from __future__ import print_function
from __future__ import absolute_import
from future import standard_library
standard_library.install_aliases()
from emspy.query import *
from .query import Query

import sys
import pickle
import warnings
import pandas as pd
import numpy as np


class TSeriesQuery(Query):


    def __init__(self, conn, ems_name, data_file = LocalData.default_data_file):

        Query.__init__(self, conn, ems_name)
        self._init_assets(data_file)
        self.reset()


    def _init_assets(self, data_file):

        # Query._init_assets(self)
        self.__analytic = Analytic(self._conn, self._ems_id, data_file)


    def reset(self):

        self.__columns  = list()
        self.__queryset = {'select':[]}


    def select_ids(self, analytic_ids, names=None, descriptions=None, units=None, lookup=False):
        """
        A method for selecting parameters to query for by analytic_id.  This method adds analytics and relevant
        information as provided.  It can also query the EMS API for relevant information if requested.

        Args:
            analytic_ids (:obj:`list` of :obj:`str`): A list of analytic_ids to select and query for.
            names (:obj:`list` of :obj:`str`): A list of names for each analytic_id in analytic_ids.

            descriptions (:obj:`list` of :obj:`str`, optional): A list of descriptions for each analytic_id in
                analytic_ids.
            units (:obj:`list` of :obj:`str`, optional): A list of units for each analytic_id in analytic_ids.
            names (:obj:`list` of :obj:`str`, optional): A list of names for each analytic_id in analytic_ids.
            lookup (:obj:`bool`, optional): Whether or not to look up names, descriptions and units using the EMS API.
        Raises:
            ValueError: If `analytic_ids`, `names`, `descriptions` (if specified), and `units` (if specified) are not
                of the same length.
            ValueError: If `analytic_ids`, and `names` are not both defined, unless `lookup=True` is specified, in which
                case names can be None.
            TypeError: If `analytic_ids` is a string, but `names`, `descriptions`, and `units` are specified as
                something other than strings.
        """

        # We require names if we are not going to look up metadata about the analytic_ids.
        # We could just allow adding analytic_ids, but the user will likely have to pair the results up with names
        # information anyway, so it might as well be provided to the select_ids method for a more friendly experience.
        if names is None and not lookup:
            raise ValueError("If `names` is not specified, you must set `lookup`=True")

        # If the input parameters are both strings, then we will cast them into lists so we can proceed as normal
        # (i.e. loop through them [once])
        if isinstance(analytic_ids, str):
            analytic_ids = [analytic_ids]
            # if names is not None, and names is not a string type object, raise an error.
            if names is not None:
                if isinstance(names, str):
                    names = [names]
                else:
                    raise TypeError("If a string is passed for `analytic_ids`, `names` must be a string as well.")

            # if descriptions is not None, and descriptions is not a string type object, raise an error.
            if descriptions is not None:
                if isinstance(descriptions, str):
                    descriptions = [descriptions]
                else:
                    raise ValueError("If a string is passed for `analytic_ids`, `descriptions` must be a string as "
                                     "well.")

            # if units is not None, and units is not a string type object, raise an error.
            if units is not None:
                if isinstance(units, str):
                    units = [units]
                else:
                    raise ValueError("If a string is passed for `analytic_ids`, `units` must be a string as well.")

        # calculate the length of each input list
        # descriptions and units aren't necessarily input lists (could be None), but they are necessarily the same
        # length as analytic_ids
        lengths = {'analytic_ids': len(analytic_ids)}
        if names is not None:
            lengths['names'] = len(names)
        if descriptions is not None:
            lengths['descriptions'] = len(descriptions)
        if units is not None:
            lengths['units'] = len(units)

        length_set = set(lengths.values())  # This is a set of unique lengths

        # If there is more than one unique iterable length, then one must be wrong.  Return an error.
        if len(length_set) > 1:
            error_str = ''
            for list_name, list_length in lengths.items():
                error_str = error_str + '\t\t{0:15}: {1}\n'.format(list_name, str(list_length))
            raise ValueError("All lists must be of the same length.  Found lengths:\n{0}".format(error_str))

        # Fill out lists with empty strings that are the same length as the analytic_ids list, if parameters are None.
        names = ['']*len(analytic_ids) if names is None else names
        descriptions = ['']*len(analytic_ids) if descriptions is None else descriptions
        units = ['']*len(analytic_ids) if units is None else units

        # loop over items
        for analytic_id, name, description, unit in zip(analytic_ids, names, descriptions, units):
            # if we aren't performing a lookup, just stuff in the lists that have been passed in.
            if not lookup:
                prm = {'id': analytic_id, 'name': name, 'description': description, 'units': unit,
                       'ems_id': self._ems_id}
            # if we are performing a lookup, we need to make an API call to get relevant data.
            else:
                prm = self.__analytic.get_param_details(analytic_id)
                prm['ems_id'] = self._ems_id
            df = pd.DataFrame(prm, index=[0])

            self.__analytic._param_table = self.__analytic._param_table.append(df, ignore_index=True, sort=True)

            # Put the param into JSON query string
            self.__queryset['select'].append({'analyticId': prm['id']})
            # Just in case you want to check what params are selected
            self.__columns.append(prm)

            self.__analytic._save_paramtable()


    def select(self, *args):
        """
        A method for selecting parameters to query for by name.  This method searches for parameters by name and adds
        them to the current query.  If multiple matches for a given parameter are found, the match with the shortest
        name is selected.

        Args:
            *args (:obj:`list` of :obj:`str`): Variable length argument list of parameter names.
        """

        keywords   = args
        save_table = False

        for kw in keywords:
            # Get the param from param table
            prm = self.__analytic.get_param(kw)
            if prm['id'] == "":
                # If the param's not found, call EMS API
                res_df = self.__analytic.search_param(kw, in_df = True)
                res_df['ems_id'] = self._ems_id
                # The first one is with the shortest name string. Pick that.
                prm = res_df.iloc[0,:].to_dict()
                # Add the new parameters to the param table for later uses
                self.__analytic._param_table = self.__analytic._param_table.append(res_df, ignore_index = True, sort = True)
                save_table = True

            # Put the param into JSON query string
            self.__queryset['select'].append({'analyticId': prm['id']})
            # Just in case you want to check what params are selected
            self.__columns.append(prm)
        if save_table:
            self.__analytic._save_paramtable()


    def range(self, start = None, end = None):

        self.__queryset['start'] = start
        self.__queryset['end']   = end


    def timepoint(self, tpoint):
        if type(tpoint) == np.ndarray:
            tpoint = tpoint.tolist()
        self.__queryset['offsets'] = tpoint


    def run(self, flight, start = None, end = None, timestep = None, timepoint = None,
            discretes_as_strings=True):

        # if start is None:
        #     start = 0.0

        # if end is None:
        #     end = self.flight_duration(flight)

        # if timestep is not None:
        #     timepoint = np.arange(start, end, timestep)

        # if timepoint is not None:
        #     self.timepoint(timepoint)
        # else:
        #     self.range(start, end)
        if timepoint is not None:
            self.timepoint(timepoint)

        # allow the user to specify discretesAsStrings = 'false'
        # per the EMS API documentation, discretesAsStrings is true by default.  So it will only get set to false if
        # that option is specified.  This will allow users to return discretes in time series parameters as int values.
        if not discretes_as_strings:
            self.__queryset['discretesAsStrings'] = 'false'

        elif timestep is not None:
            start = 0.0 if start is None else start

            if end is None:
                raise ValueError("End timepoint should be given along with timestep input.")

            self.timepoint(np.arange(start, end+1e-10, timestep))

        else:
            self.range(start, end)

        resp_h, content = self._conn.request( uri_keys = ("analytic", "query"),
                                              uri_args = (self._ems_id, flight),
                                              jsondata = self.__queryset)

        if 'message' in content:
            sys.exit('API query for flight %d was unsuccessful.\nHere is the message from API: %s' % (flight, content['message']))
        
        # Put the data in Pandas DataFrame
        df = pd.DataFrame({"Time (sec)": content['offsets']})

        for i, prm in enumerate(self.__columns):
            df[prm['name']] = content['results'][i]['values']

        ## "\r\x1b[K" is overwrite print.
        return df



    def multi_run(self, flight, start = None, end = None, timestep=None, timepoint = None, save_file = None, verbose = True):

        res       = list()
        attr_flag = False
        if isinstance(flight, pd.DataFrame):
            FR = flight["Flight Record"]
            attr_flag = True
        else:
            FR = flight

        # param processing
        if start is None:       start       = [None]*len(FR)
        if end   is None:       end         = [None]*len(FR)
        if not hasattr(timestep, "__len__"):
            timestep = [timestep]*len(FR)
        if timepoint is None:   
            timepoint   = [None]*len(FR)
        else: 
            warnings.warn("Time points are not yet supported. The given time points will be ignored.")
            timepoint   = [None]*len(FR)
            
        if verbose: print('\n=== Start running time-series data querying for %d flights ===\n' % len(FR))
        
        for i, fr in enumerate(FR):
            if verbose: print('\r\x1b[K%d / %d: FR %d' % (i+1, len(FR), fr), end=' ')
            i_res = dict()
            if attr_flag:
                i_res['flt_data'] = flight.iloc[i,:].to_dict()
            else:
                i_res['flt_data'] = {'Flight Record': fr}
            i_res['ts_data'] = self.run(fr, start[i], end[i], timestep[i])
            res.append(i_res)

            if save_file is not None:
                pickle.dump(res, open(save_file, 'wb'))
        if verbose: print('Done')

        return res


    def flight_duration(self, flight, unit = "second"):
        '''
        deprecated
        '''

        p = self.__analytic.get_param("hours of data (hours)")
        if p["id"] == "":
            res_df = self.__analytic.search_param("hours of data (hours)", in_df = True)
            p      = res_df.iloc[0].to_dict()
            self.__analytic._param_table = self.__analytic._param_table.append(res_df, ignore_index = True, sort = True)
            self.__analytic._save_paramtable()
        q = {
            "select": [{"analyticId": p["id"]}],
            "size": 1
        }

        resp_h, content = self._conn.request( uri_keys = ("analytic", "query"),
                                              uri_args = (self._ems_id, flight),
                                              jsondata = q)
        if 'message' in content:
            sys.exit('API query for flight %d, parameter = "%s" was unsuccessful.\nHere is the message from API: %s' % (flight, p['name'], content['message']))
        fl_len = content['results'][0]['values'][0]

        if unit == "second":
            t = fl_len * 60 * 60
        elif unit == "minute":
            t = fl_len * 60
        elif unit == "hour":
            t = fl_len
        else:
            sys.exit("Unrecognizable time unit (%s)." % unit)
        return t

    def get_queryset(self):
        """
        Returns a dictionary with the queryset for the current time series query.

        Returns:
            list: a list describing the current queryset.
        """

        return self.__queryset




from emspy.query.ems import EMS
from emspy.query.fleet import Fleet
from emspy.query.aircraft import Aircraft
from emspy.query.airport import Airport
from emspy.query.flightphase import FlightPhase
from emspy.query.localdata import LocalData
from emspy.query.flight import Flight
from emspy.query.analytic import Analytic
from emspy.query.fltquery import FltQuery
from emspy.query.tsquery import TSeriesQuery
from emspy.query.profile import Profile
from emspy.query.dbmodify import InsertQuery




from builtins import object
import emspy

import pandas as pd
import numpy as np
import os, sys, re, sqlite3


class LocalData(object):
	default_data_file = os.path.join(emspy.__path__[0], "data","emsMetaData.db")

	table_info = {
		"fieldtree": ["ems_id", "db_id", "id", "nodetype", "type", "name", "parent_id" ],
		"dbtree"   : ["ems_id", "id", "nodetype", "name", "parent_id"],
		"kvmaps"   : ["ems_id", "id", "key", "value"],
		"params"   : ["ems_id", "id", "name", "description", "units"]
	}

	def __init__(self, dbfile = default_data_file):

		self.__dbfile = None
		self._conn = None

		# This is done to support a mode that does not use a data file at all (when dbfile is set to None explicitly)
		if dbfile:
			self.__dbfile = os.path.abspath(dbfile)
			self.__connect()

	
	def __del__(self):

		self.close()


	def __connect(self):

		if self.__dbfile:
			self._conn = sqlite3.connect(self.__dbfile)


	def __check_colnames(self, table_name, df):

		colnames = np.array(LocalData.table_info[table_name])
		chk_cols = np.array([c in df.columns for c in colnames])
		missing  = colnames[~chk_cols]
		if any(~chk_cols):
			sys.exit("Data misses the following columns that are required: %s" % missing)


	def close(self):

		if self._conn is not None:
			self._conn.close()


	def append_data(self, table_name, df):
		
		self.__check_colnames(table_name, df)
		if self.__dbfile is not None:
			df.to_sql(table_name, self._conn, index=False, if_exists="append")


	def get_data(self, table_name, condition = None):

		if (self.__dbfile is not None) and self.table_exists(table_name):
			q  = "SELECT * FROM %s" % table_name
			if condition is not None:
				q = q + " WHERE %s" % condition
			q  = q + ";"
			df = pd.read_sql_query(q, self._conn)

			# Strange columns appear. Get only the actual columns
			return df[LocalData.table_info[table_name]]		
		return pd.DataFrame(columns = LocalData.table_info[table_name])


	def delete_data(self, table_name, condition = None):

		if (self.__dbfile is not None) and self.table_exists(table_name):
			if condition is None:
				self._conn.execute("DROP TABLE %s" % table_name)
			else:
				self._conn.execute("DELETE FROM %s WHERE %s;" % (table_name, condition))
			self._conn.commit()


	def delete_all_tables(self):

		if self.__dbfile is not None:
			for table_name in list(LocalData.table_info.keys()):
				if self.table_exists(table_name):
					self._conn.execute("DROP TABLE %s" % table_name)
			self._conn.commit()


	def table_exists(self, table_name):

		if self.__dbfile is None:
			return False

		cursor = self._conn.cursor()
		cursor.execute("SELECT name FROM sqlite_master WHERE type='table' ORDER BY name")
		tables = [t[0] for t in cursor.fetchall()]
		return table_name in tables


	def file_loc(self):

		return self.__dbfile

	def is_db_path_correct(self, path):
		"""Checks if the path represents the current data file.
		
		Arguments:
			path (str): The path to compare.
		
		Returns:
			bool: True if the path is the current data file, false otherwise.
		"""

		if path is None:
			return self.file_loc() == None
		else:
			return self.file_loc() == os.path.abspath(path)





# uri_root = 'https://fas.efoqa.com/api'
# uri_root = 'https://ceod.efoqa.com/api'
uri_root = {
	'prod'		: 'https://ems.efoqa.com/api',
	'cluster'	: 'https://ceod.efoqa.com/api',
	'stable'	: 'https://emsapi.ausdig.com/api',
	'beta'		: 'https://emsapibeta.ausdig.com/api',
	'nightly'	: 'https://emsapitest.ausdig.com/api'
}

user_agent = 'ems-api-sdk Python v0.2'

uris = {
	'sys': {
		'auth': '/token'
	},
	'ems_sys': {
		'list': '/v2/ems-systems',
		'ping': '/v2/ems-systems/%s/ping', 	# (ems-system_id)
		'info': '/v2/ems-systems/%s/info'	# (ems-system_id)	
	},
	'fleet': {
		'list'	: '/v2/ems-systems/%s/assets/fleets', 	#(ems-system_id)
		'info'	: '/v2/ems-systems/%s/assets/fleets/%s'	#(ems-system_id, fleet_id)	
	},
	'aircraft': {
		'list'	: '/v2/ems-systems/%s/assets/aircraft',	#(ems-system_id)
		'info'	: '/v2/ems-systems/%s/assets/aircraft/%s'#(ems-system_id, aircraft_id)
	},
	'flt_phase': {
		'list'	: '/v2/ems-systems/%s/assets/flight-phases', 	#(ems-system_id)
		'info'	: '/v2/ems-systems/%s/assets/flight-phases/%s' 	#(ems-system_id, flt_phase_id)
	},
	'airport': {
		'list'	: '/v2/ems-systems/%s/assets/airports',		#(ems-system_id)
		'info'	: '/v2/ems-systems/%s/assets/airports/%s'	#(ems-system_id, airport_id)
	},
	'database': {
		'group'			: '/v2/ems-systems/%s/database-groups',			#(ems-system_id)
		'field_group'	: '/v2/ems-systems/%s/databases/%s/field-groups',	#(ems-system_id, data_src_id)
		'field'			: '/v2/ems-systems/%s/databases/%s/fields/%s', #(ems-system_id, database_id, field_id)
		'query'			: '/v2/ems-systems/%s/databases/%s/query',
		'create'		: '/v2/ems-systems/%s/databases/%s/create', # (emsSystemId, databaseId)
		'open_asyncq'	: '/v2/ems-systems/%s/databases/%s/async-query', #(ems-system_id, database_id)
		'get_asyncq'	: '/v2/ems-systems/%s/databases/%s/async-query/%s/read/%s/%s', #(ems-system_id, database_id, async_query_id, start_row, end_row)
		'close_asyncq'	: '/v2/ems-systems/%s/databases/%s/async-query/%s'
	},
	'analytic': {
		'search'	: '/v2/ems-systems/%s/analytics',    # (emsSystemId)
		'search_f'	: '/v2/ems-systems/%s/flights/%s/analytics', # (emsSystemId, flightId)
		'group'		: '/v2/ems-systems/%s/analytic-groups',    # (emsSystemId)
		'group_f'	: '/v2/ems-systems/%s/flights/%s/analytic-groups', # (emsSystemId, flightId)
		'query'		: '/v2/ems-systems/%s/flights/%s/analytics/query', # (emsSystemId, flightId)
		'metadata'	: '/v2/ems-systems/%s/flights/%s/analytics/metadata' # (emsSystemId, flightId)
	},
	'profile': {
		'profile_results'	: '/v2/ems-systems/%s/flights/%s/profiles/%s/query',  # (emsSystemId, flightId, profileId)
		'search'			: '/v2/ems-systems/%s/profiles',  # (emsSystemId)
		'glossary'			: '/v2/ems-systems/%s/profiles/%s/glossary',  # (emsSystemId, profileId)
		'events'			: '/v2/ems-systems/%s/profiles/%s/events',  # (emsSystemId, profileId)
		'single_event'		: '/v2/ems-systems/%s/profiles/%s/events/%s',  # (emsSystemId, profileId, eventId)
	}

}



import sys
from setuptools import setup, find_packages

def readme():
    with open('README.md') as f:
        return f.read()

# Numpy dropped support for python 2.7 in 1.17
# Pandas dropped support for python 2.7 after 0.24.0
min_numpy_ver = "1.13.3"
max_numpy_ver = ""
pandas_install_rule = 'pandas'
if sys.version_info < (3, 0):
    max_numpy_ver = "1.17"
    pandas_install_rule += "<=0.24.0"
    
numpy_install_rule = "numpy >= {numpy_ver}".format(numpy_ver=min_numpy_ver)
if max_numpy_ver:
    numpy_install_rule += ", < {numpy_ver}".format(numpy_ver=max_numpy_ver)

setup(name = 'emspy',
      version = '0.2.6',
      description = 'A Python EMS RESTful API Client/Wrapper',
      long_description=readme(),
      long_description_content_type="text/markdown",
      classifiers = [
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License'
        ],
      url = 'https://github.com/ge-flight-analytics/emspy',
      author = 'GE Flight Analytics',
      author_email = 'AviationAdiSupport@ge.com',
      license = 'MIT',
      packages = find_packages(exclude=("docs","tests")),
      install_requires = [numpy_install_rule, pandas_install_rule, 'future', 'networkx'],
      zip_safe = False)
      


from builtins import object
import pandas as pd


class Asset(object):
	'''
	Super class of all "asset" classes and the EMS class. It is a template class so 
	should not be instantiated directly.
	'''

	def __init__(self, conn, asset_type):

		self._conn = conn
		self._asset_type = asset_type
		self._assets = None


	def update_list(self, uri_keys=None, uri_args=None, body=None, colsort=True):

		resp_h, dict_data = self._conn.request(uri_keys=uri_keys, uri_args=uri_args, body=body)
		a = pd.DataFrame.from_dict(dict_data)
		if colsort:
			a = a[sorted(a.columns, key=len)]
		self._assets = a


	def list_all(self):

		return self._assets


	def search(self, col, val = None, searchtype="contain"):

		if val is None:
			return self.list_all()
		
		a = self._assets

		if type(val)==str:			
			if searchtype=="contain":
				return a[a[col].str.contains(val, na=False, case=False)]
			if searchtype=="match":
				return a[a[col].str.match(val, case=False)]
		
		return a[a[col]==val]


	def get_id(self, name = None, keyword = None):

		raise NotImplementedError()


	def get_name(self, id_val = None):

		raise NotImplementedError()
		

	def _rename_datacol(self, old, new):

		self._assets = self._assets.rename(columns = {old:new})


	def data_colnames(self):

		return self._assets.columns


def size(an_asset):
	
	if not issubclass(an_asset, Asset):
		raise TypeError("Input should be an object inherited from Asset class.")

	if an_asset.list_all() is None: 
		return 0
	else:
		return an_asset.list_all().shape[0]



